__title__ = 'compas_nurbs'
__description__ = 'NURBS for COMPAS'
__url__ = 'https://github.com/gramaziokohler/compas_nurbs'
__version__ = '0.1.0'
__author__ = 'Gramazio Kohler Research'
__author_email__ = 'gramaziokohler@arch.ethz.ch'
__license__ = 'MIT license'
__copyright__ = 'Copyright 2020 ETH Zurich, Gramazio Kohler Research'

__all__ = ['__author__', '__author_email__', '__copyright__', '__description__', '__license__', '__title__', '__url__', '__version__']
